"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var immutable_1 = require("immutable");
var pluralize = require("pluralize");
var Type_1 = require("./Type");
var Support_1 = require("./Support");
var TypeNames_1 = require("./TypeNames");
var TypeNames_2 = require("./TypeNames");
var TypeAttributes_1 = require("./TypeAttributes");
var PathElementKind;
(function (PathElementKind) {
    PathElementKind[PathElementKind["KeyOrIndex"] = 0] = "KeyOrIndex";
    PathElementKind[PathElementKind["Type"] = 1] = "Type";
    PathElementKind[PathElementKind["Object"] = 2] = "Object";
})(PathElementKind || (PathElementKind = {}));
function keyOrIndex(pe) {
    if (pe.kind !== PathElementKind.KeyOrIndex)
        return undefined;
    return pe.key;
}
function pathElementEquals(a, b) {
    if (a.kind !== b.kind)
        return false;
    switch (a.kind) {
        case PathElementKind.Type:
            return a.index === b.index;
        case PathElementKind.KeyOrIndex:
            return a.key === b.key;
        default:
            return true;
    }
}
var numberRegexp = new RegExp("^[0-9]+$");
var Ref = /** @class */ (function () {
    function Ref(_path) {
        this._path = _path;
    }
    Ref.parse = function (ref) {
        if (typeof ref !== "string") {
            return Support_1.panic("$ref must be a string");
        }
        if (!ref.startsWith("#/")) {
            return Support_1.panic('$ref must start with "#/"');
        }
        ref = ref.substr(2);
        if (ref === "")
            return Ref.root;
        var elements = [];
        var parts = ref.split("/");
        for (var i = 0; i < parts.length; i++) {
            elements.push({ kind: PathElementKind.KeyOrIndex, key: parts[i] });
        }
        return new Ref(immutable_1.List(elements));
    };
    Ref.prototype.pushElement = function (pe) {
        return new Ref(this._path.push(pe));
    };
    Ref.prototype.push = function () {
        var keys = [];
        for (var _i = 0; _i < arguments.length; _i++) {
            keys[_i] = arguments[_i];
        }
        var ref = this;
        for (var _a = 0, keys_1 = keys; _a < keys_1.length; _a++) {
            var key = keys_1[_a];
            ref = ref.pushElement({ kind: PathElementKind.KeyOrIndex, key: key });
        }
        return ref;
    };
    Ref.prototype.pushObject = function () {
        return this.pushElement({ kind: PathElementKind.Object });
    };
    Ref.prototype.pushType = function (index) {
        return this.pushElement({ kind: PathElementKind.Type, index: index });
    };
    Object.defineProperty(Ref.prototype, "name", {
        get: function () {
            var path = this._path;
            for (;;) {
                var e = path.last();
                if (e === undefined) {
                    return "Something";
                }
                switch (e.kind) {
                    case PathElementKind.KeyOrIndex:
                        if (e.key.match(numberRegexp) !== null) {
                            return e.key;
                        }
                        break;
                    case PathElementKind.Type:
                    case PathElementKind.Object:
                        return Support_1.panic("We shouldn't try to get the name of Type or Object refs");
                    default:
                        return Support_1.assertNever(e);
                }
                path = path.pop();
            }
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Ref.prototype, "definitionName", {
        get: function () {
            var pe = this._path.get(-2);
            if (pe === undefined)
                return undefined;
            if (keyOrIndex(pe) === "definitions")
                return keyOrIndex(Support_1.defined(this._path.last()));
            return undefined;
        },
        enumerable: true,
        configurable: true
    });
    Ref.prototype.toString = function () {
        function elementToString(e) {
            switch (e.kind) {
                case PathElementKind.Type:
                    return "type/" + e.index.toString();
                case PathElementKind.Object:
                    return "object";
                case PathElementKind.KeyOrIndex:
                    return e.key;
                default:
                    return Support_1.assertNever(e);
            }
        }
        return "#/" + this._path.map(elementToString).join("/");
    };
    Ref.prototype.lookupRef = function (root) {
        function lookup(local, path) {
            var first = path.first();
            if (first === undefined) {
                return Support_1.checkStringMap(local);
            }
            var rest = path.rest();
            switch (first.kind) {
                case PathElementKind.KeyOrIndex:
                    if (Array.isArray(local)) {
                        return lookup(local[parseInt(first.key, 10)], rest);
                    }
                    else {
                        return lookup(Support_1.checkStringMap(local)[first.key], rest);
                    }
                case PathElementKind.Type:
                    return Support_1.panic('Cannot look up path that indexes "type"');
                case PathElementKind.Object:
                    return Support_1.panic('Cannot look up path that indexes "object"');
                default:
                    return Support_1.assertNever(first);
            }
        }
        return lookup(root, this._path);
    };
    Ref.prototype.equals = function (other) {
        if (!(other instanceof Ref))
            return false;
        if (this._path.size !== other._path.size)
            return false;
        return this._path.zipWith(pathElementEquals, other._path).every(function (x) { return x; });
    };
    Ref.prototype.hashCode = function () {
        var acc = Support_1.hashCodeInit;
        this._path.forEach(function (pe) {
            acc = Support_1.addHashCode(acc, pe.kind);
            switch (pe.kind) {
                case PathElementKind.Type:
                    acc = Support_1.addHashCode(acc, pe.index);
                    break;
                case PathElementKind.KeyOrIndex:
                    acc = Support_1.addHashCode(acc, immutable_1.hash(pe.key));
                    break;
            }
        });
        return acc;
    };
    Ref.root = new Ref(immutable_1.List());
    return Ref;
}());
exports.Ref = Ref;
function checkStringArray(arr) {
    if (!Array.isArray(arr)) {
        return Support_1.panic("Expected a string array, but got " + arr);
    }
    for (var _i = 0, arr_1 = arr; _i < arr_1.length; _i++) {
        var e = arr_1[_i];
        if (typeof e !== "string") {
            return Support_1.panic("Expected string, but got " + e);
        }
    }
    return arr;
}
function makeAttributes(schema, path, attributes) {
    var maybeDescription = schema.description;
    if (typeof maybeDescription === "string") {
        attributes = TypeAttributes_1.descriptionTypeAttributeKind.setInAttributes(attributes, immutable_1.OrderedSet([maybeDescription]));
    }
    return TypeNames_2.modifyTypeNames(attributes, function (maybeTypeNames) {
        var typeNames = Support_1.defined(maybeTypeNames);
        if (!typeNames.areInferred) {
            return typeNames;
        }
        var title = schema.title;
        if (typeof title !== "string") {
            title = path.definitionName;
        }
        if (typeof title === "string") {
            return new TypeNames_1.TypeNames(immutable_1.OrderedSet([title]), immutable_1.OrderedSet(), false);
        }
        else {
            return typeNames.makeInferred();
        }
    });
}
function checkTypeList(typeOrTypes) {
    if (typeof typeOrTypes === "string") {
        return immutable_1.OrderedSet([typeOrTypes]);
    }
    else if (Array.isArray(typeOrTypes)) {
        var arr = [];
        for (var _i = 0, typeOrTypes_1 = typeOrTypes; _i < typeOrTypes_1.length; _i++) {
            var t = typeOrTypes_1[_i];
            if (typeof t !== "string") {
                return Support_1.panic("element of type is not a string: " + t);
            }
            arr.push(t);
        }
        var set = immutable_1.OrderedSet(arr);
        Support_1.assert(!set.isEmpty(), "JSON Schema must specify at least one type");
        return set;
    }
    else {
        return Support_1.panic("type is neither a string or array of strings: " + typeOrTypes);
    }
}
function addTypesInSchema(typeBuilder, rootJson, references) {
    var root = Support_1.checkStringMap(rootJson);
    var typeForPath = immutable_1.Map();
    function setTypeForPath(path, t) {
        var maybeRef = typeForPath.get(path);
        if (maybeRef !== undefined) {
            Support_1.assert(maybeRef === t, "Trying to set path again to different type");
        }
        typeForPath = typeForPath.set(path, t);
    }
    function makeClass(path, attributes, properties, requiredArray) {
        var required = immutable_1.Set(requiredArray);
        var propertiesMap = immutable_1.Map(properties);
        var propertyDescriptions = propertiesMap
            .map(function (propSchema) {
            if (typeof propSchema === "object") {
                var desc = propSchema.description;
                if (typeof desc === "string") {
                    return immutable_1.OrderedSet([desc]);
                }
            }
            return undefined;
        })
            .filter(function (v) { return v !== undefined; });
        if (!propertyDescriptions.isEmpty()) {
            attributes = TypeAttributes_1.propertyDescriptionsTypeAttributeKind.setInAttributes(attributes, propertyDescriptions);
        }
        // FIXME: We're using a Map instead of an OrderedMap here because we represent
        // the JSON Schema as a JavaScript object, which has no map ordering.  Ideally
        // we would use a JSON parser that preserves order.
        var props = propertiesMap.map(function (propSchema, propName) {
            var t = toType(Support_1.checkStringMap(propSchema), path.push("properties", propName), TypeNames_2.makeNamesTypeAttributes(pluralize.singular(propName), true));
            var isOptional = !required.has(propName);
            return new Type_1.ClassProperty(t, isOptional);
        });
        return typeBuilder.getUniqueClassType(attributes, true, props.toOrderedMap());
    }
    function makeMap(path, typeAttributes, additional) {
        path = path.push("additionalProperties");
        var valuesType = toType(additional, path, TypeNames_2.singularizeTypeNames(typeAttributes));
        return typeBuilder.getMapType(valuesType);
    }
    function fromTypeName(schema, path, typeAttributes, typeName) {
        // FIXME: We seem to be overzealous in making attributes.  We get them from
        // our caller, then we make them again here, and then we make them again
        // in `makeClass`, potentially in other places, too.
        typeAttributes = makeAttributes(schema, path, TypeAttributes_1.makeTypeAttributesInferred(typeAttributes));
        switch (typeName) {
            case "object":
                var required = void 0;
                if (schema.required === undefined) {
                    required = [];
                }
                else {
                    required = checkStringArray(schema.required);
                }
                // FIXME: Don't put type attributes in the union AND its members.
                var unionType = typeBuilder.getUniqueUnionType(typeAttributes, undefined);
                setTypeForPath(path, unionType);
                var typesInUnion = [];
                if (schema.properties !== undefined) {
                    typesInUnion.push(makeClass(path, typeAttributes, Support_1.checkStringMap(schema.properties), required));
                }
                if (schema.additionalProperties !== undefined) {
                    var additional = schema.additionalProperties;
                    // FIXME: We don't treat `additional === true`, which is also the default,
                    // not according to spec.  It should be translated into a map type to any,
                    // though that's not what the intention usually is.  Ideally, we'd find a
                    // way to store additional attributes on regular classes.
                    if (additional === false) {
                        if (schema.properties === undefined) {
                            typesInUnion.push(makeClass(path, typeAttributes, {}, required));
                        }
                    }
                    else if (typeof additional === "object") {
                        typesInUnion.push(makeMap(path, typeAttributes, Support_1.checkStringMap(additional)));
                    }
                }
                if (typesInUnion.length === 0) {
                    typesInUnion.push(typeBuilder.getMapType(typeBuilder.getPrimitiveType("any")));
                }
                typeBuilder.setSetOperationMembers(unionType, immutable_1.OrderedSet(typesInUnion));
                return unionType;
            case "array":
                if (schema.items !== undefined) {
                    path = path.push("items");
                    return typeBuilder.getArrayType(toType(Support_1.checkStringMap(schema.items), path, TypeNames_2.singularizeTypeNames(typeAttributes)));
                }
                return typeBuilder.getArrayType(typeBuilder.getPrimitiveType("any"));
            case "boolean":
                return typeBuilder.getPrimitiveType("bool");
            case "string":
                if (schema.format !== undefined) {
                    switch (schema.format) {
                        case "date":
                            return typeBuilder.getPrimitiveType("date");
                        case "time":
                            return typeBuilder.getPrimitiveType("time");
                        case "date-time":
                            return typeBuilder.getPrimitiveType("date-time");
                        default:
                            // FIXME: Output a warning here instead to indicate that
                            // the format is uninterpreted.
                            return typeBuilder.getStringType(typeAttributes, undefined);
                    }
                }
                return typeBuilder.getStringType(typeAttributes, undefined);
            case "null":
                return typeBuilder.getPrimitiveType("null");
            case "integer":
                return typeBuilder.getPrimitiveType("integer");
            case "number":
                return typeBuilder.getPrimitiveType("double");
            default:
                return Support_1.panic("not a type name: " + typeName);
        }
    }
    function convertToType(schema, path, typeAttributes) {
        typeAttributes = makeAttributes(schema, path, typeAttributes);
        function makeTypesFromCases(cases, kind) {
            if (!Array.isArray(cases)) {
                return Support_1.panic("Cases are not an array: " + cases);
            }
            // FIXME: This cast shouldn't be necessary, but TypeScript forces our hand.
            return cases.map(function (t, index) {
                return toType(Support_1.checkStringMap(t), path.push(kind, index.toString()), TypeAttributes_1.makeTypeAttributesInferred(typeAttributes));
            });
        }
        function convertOneOrAnyOf(cases, kind) {
            var unionType = typeBuilder.getUniqueUnionType(TypeAttributes_1.makeTypeAttributesInferred(typeAttributes), undefined);
            typeBuilder.setSetOperationMembers(unionType, immutable_1.OrderedSet(makeTypesFromCases(cases, kind)));
            return unionType;
        }
        if (schema.$ref !== undefined) {
            var ref_1 = Ref.parse(schema.$ref);
            var target = ref_1.lookupRef(root);
            var attributes = TypeNames_2.modifyTypeNames(typeAttributes, function (tn) {
                if (!Support_1.defined(tn).areInferred)
                    return tn;
                return new TypeNames_1.TypeNames(immutable_1.OrderedSet([ref_1.name]), immutable_1.OrderedSet(), true);
            });
            return toType(target, ref_1, attributes);
        }
        else if (Array.isArray(schema.enum)) {
            var cases = schema.enum;
            var haveNull = cases.indexOf(null) >= 0;
            cases = cases.filter(function (c) { return c !== null; });
            if (cases.filter(function (c) { return typeof c !== "string"; }).length > 0) {
                return Support_1.panic("Non-string enum cases are not supported, at " + path.toString());
            }
            var tref = typeBuilder.getEnumType(typeAttributes, immutable_1.OrderedSet(checkStringArray(cases)));
            if (haveNull) {
                return typeBuilder.getUnionType(typeAttributes, immutable_1.OrderedSet([tref, typeBuilder.getPrimitiveType("null")]));
            }
            else {
                return tref;
            }
        }
        var jsonTypes = undefined;
        if (schema.type !== undefined) {
            jsonTypes = checkTypeList(schema.type);
        }
        else if (schema.properties !== undefined || schema.additionalProperties !== undefined) {
            jsonTypes = immutable_1.OrderedSet(["object"]);
        }
        var intersectionType = typeBuilder.getUniqueIntersectionType(typeAttributes, undefined);
        setTypeForPath(path, intersectionType);
        var types = [];
        if (schema.allOf !== undefined) {
            types.push.apply(types, makeTypesFromCases(schema.allOf, "allOf"));
        }
        if (schema.oneOf) {
            types.push(convertOneOrAnyOf(schema.oneOf, "oneOf"));
        }
        if (schema.anyOf) {
            types.push(convertOneOrAnyOf(schema.anyOf, "anyOf"));
        }
        if (jsonTypes !== undefined) {
            if (jsonTypes.size === 1) {
                types.push(fromTypeName(schema, path.pushObject(), typeAttributes, Support_1.defined(jsonTypes.first())));
            }
            else {
                var unionType = typeBuilder.getUniqueUnionType(typeAttributes, undefined);
                var unionTypes = jsonTypes
                    .toList()
                    .map(function (n, index) { return fromTypeName(schema, path.pushType(index), typeAttributes, n); });
                typeBuilder.setSetOperationMembers(unionType, immutable_1.OrderedSet(unionTypes));
                types.push(unionType);
            }
        }
        typeBuilder.setSetOperationMembers(intersectionType, immutable_1.OrderedSet(types));
        return intersectionType;
    }
    function toType(schema, path, typeAttributes) {
        var maybeType = typeForPath.get(path);
        if (maybeType !== undefined) {
            return maybeType;
        }
        var result = convertToType(schema, path, typeAttributes);
        setTypeForPath(path, result);
        return result;
    }
    references.forEach(function (topLevelRef, topLevelName) {
        var target = topLevelRef.lookupRef(root);
        var t = toType(target, topLevelRef, TypeNames_2.makeNamesTypeAttributes(topLevelName, false));
        typeBuilder.addTopLevel(topLevelName, t);
    });
}
exports.addTypesInSchema = addTypesInSchema;
function definitionRefsInSchema(rootJson) {
    if (typeof rootJson !== "object")
        return immutable_1.Map();
    var definitions = rootJson.definitions;
    if (typeof definitions !== "object")
        return immutable_1.Map();
    return immutable_1.Map(Object.keys(definitions).map(function (name) {
        return [name, Ref.root.push("definitions", name)];
    }));
}
exports.definitionRefsInSchema = definitionRefsInSchema;
