import { Map, OrderedMap, OrderedSet, Set } from "immutable";
import { PrimitiveTypeKind, Type, PrimitiveStringTypeKind, ClassProperty, TypeKind } from "./Type";
import { TypeGraph } from "./TypeGraph";
import { TypeAttributes, TypeAttributeKind } from "./TypeAttributes";
export declare type TypeRefCallback = (index: number) => void;
export declare class TypeRef {
    readonly graph: TypeGraph;
    private _allocatingTypeBuilder;
    private _maybeIndexOrRef?;
    private _callbacks?;
    constructor(graph: TypeGraph, index?: number, _allocatingTypeBuilder?: TypeBuilder | undefined);
    private follow();
    readonly maybeIndex: number | undefined;
    getIndex(): number;
    callWhenResolved(callback: TypeRefCallback): void;
    resolve(tref: TypeRef): void;
    deref(): [Type, TypeAttributes];
    equals(other: any): boolean;
    hashCode(): number;
}
export declare const provenanceTypeAttributeKind: TypeAttributeKind<Set<TypeRef>>;
export declare type StringTypeMapping = {
    date: PrimitiveStringTypeKind;
    time: PrimitiveStringTypeKind;
    dateTime: PrimitiveStringTypeKind;
};
export declare const NoStringTypeMapping: StringTypeMapping;
export declare abstract class TypeBuilder {
    private readonly _stringTypeMapping;
    readonly alphabetizeProperties: boolean;
    private readonly _allPropertiesOptional;
    private readonly _addProvenanceAttributes;
    readonly typeGraph: TypeGraph;
    protected topLevels: Map<string, TypeRef>;
    protected readonly types: (Type | undefined)[];
    private readonly typeAttributes;
    constructor(_stringTypeMapping: StringTypeMapping, alphabetizeProperties: boolean, _allPropertiesOptional: boolean, _addProvenanceAttributes: boolean, inheritsProvenanceAttributes: boolean);
    addTopLevel(name: string, tref: TypeRef): void;
    reserveTypeRef(): TypeRef;
    private commitType;
    protected addType<T extends Type>(forwardingRef: TypeRef | undefined, creator: (tref: TypeRef) => T, attributes: TypeAttributes | undefined): TypeRef;
    atIndex(index: number): [Type, TypeAttributes];
    addAttributes(tref: TypeRef, attributes: TypeAttributes | undefined): void;
    makeNullable(tref: TypeRef, attributes: TypeAttributes): TypeRef;
    finish(): TypeGraph;
    private _primitiveTypes;
    private _noEnumStringType;
    private _mapTypes;
    private _arrayTypes;
    private _enumTypes;
    private _classTypes;
    private _unionTypes;
    getPrimitiveType(kind: PrimitiveTypeKind, forwardingRef?: TypeRef): TypeRef;
    getStringType(attributes: TypeAttributes | undefined, cases: OrderedMap<string, number> | undefined, forwardingRef?: TypeRef): TypeRef;
    getEnumType(attributes: TypeAttributes, cases: OrderedSet<string>, forwardingRef?: TypeRef): TypeRef;
    getMapType(values: TypeRef, forwardingRef?: TypeRef): TypeRef;
    getArrayType(items: TypeRef, forwardingRef?: TypeRef): TypeRef;
    modifyPropertiesIfNecessary(properties: OrderedMap<string, ClassProperty>): OrderedMap<string, ClassProperty>;
    getClassType(attributes: TypeAttributes, properties: OrderedMap<string, ClassProperty>, forwardingRef?: TypeRef): TypeRef;
    getUniqueClassType(attributes: TypeAttributes, isFixed: boolean, properties?: OrderedMap<string, ClassProperty>, forwardingRef?: TypeRef): TypeRef;
    setClassProperties(ref: TypeRef, properties: OrderedMap<string, ClassProperty>): void;
    getUnionType(attributes: TypeAttributes, members: OrderedSet<TypeRef>, forwardingRef?: TypeRef): TypeRef;
    getUniqueUnionType(attributes: TypeAttributes, members: OrderedSet<TypeRef> | undefined, forwardingRef?: TypeRef): TypeRef;
    getUniqueIntersectionType(attributes: TypeAttributes, members: OrderedSet<TypeRef> | undefined, forwardingRef?: TypeRef): TypeRef;
    setSetOperationMembers(ref: TypeRef, members: OrderedSet<TypeRef>): void;
    abstract setLostTypeAttributes(): void;
}
export interface TypeLookerUp {
    lookupTypeRefs(typeRefs: TypeRef[], forwardingRef?: TypeRef): TypeRef | undefined;
    reconstituteTypeRef(typeRef: TypeRef, forwardingRef?: TypeRef): TypeRef;
    registerUnion(typeRefs: TypeRef[], reconstituted: TypeRef): void;
}
export declare class TypeGraphBuilder extends TypeBuilder {
    protected typeForEntry(entry: Type | undefined): Type | undefined;
    getLazyMapType(valuesCreator: () => TypeRef | undefined): TypeRef;
    setLostTypeAttributes(): void;
}
export declare class TypeReconstituter {
    private readonly _typeBuilder;
    private readonly _makeClassUnique;
    private readonly _typeAttributes;
    private readonly _forwardingRef;
    private _wasUsed;
    constructor(_typeBuilder: TypeBuilder, _makeClassUnique: boolean, _typeAttributes: TypeAttributes, _forwardingRef: TypeRef);
    private useBuilder();
    private addAttributes(tref);
    getPrimitiveType(kind: PrimitiveTypeKind): TypeRef;
    getStringType(enumCases: OrderedMap<string, number> | undefined): TypeRef;
    getEnumType(cases: OrderedSet<string>): TypeRef;
    getMapType(values: TypeRef): TypeRef;
    getArrayType(items: TypeRef): TypeRef;
    getClassType(properties: OrderedMap<string, ClassProperty>): TypeRef;
    getUniqueClassType(isFixed: boolean, properties?: OrderedMap<string, ClassProperty>): TypeRef;
    getUnionType(members: OrderedSet<TypeRef>): TypeRef;
    getUniqueIntersectionType(members: OrderedSet<TypeRef>): TypeRef;
}
export declare class GraphRewriteBuilder<T extends Type> extends TypeBuilder implements TypeLookerUp {
    private readonly _originalGraph;
    private readonly _replacer;
    private _setsToReplaceByMember;
    private _reconstitutedTypes;
    private _reconstitutedUnions;
    private _lostTypeAttributes;
    constructor(_originalGraph: TypeGraph, stringTypeMapping: StringTypeMapping, alphabetizeProperties: boolean, graphHasProvenanceAttributes: boolean, setsToReplace: T[][], _replacer: (typesToReplace: Set<T>, builder: GraphRewriteBuilder<T>, forwardingRef: TypeRef) => TypeRef);
    registerUnion(typeRefs: TypeRef[], reconstituted: TypeRef): void;
    followIndex(index: number): number;
    protected typeForEntry(entry: Type | undefined | number): Type | undefined;
    withForwardingRef(maybeForwardingRef: TypeRef | undefined, typeCreator: (forwardingRef: TypeRef) => TypeRef): TypeRef;
    private replaceSet(typesToReplace, maybeForwardingRef);
    private forceReconstituteTypeRef(originalRef, maybeForwardingRef?);
    private assertTypeRefsToReconstitute(typeRefs, forwardingRef?);
    reconstituteTypeRef(originalRef: TypeRef, maybeForwardingRef?: TypeRef): TypeRef;
    reconstituteType(t: Type, forwardingRef?: TypeRef): TypeRef;
    lookupTypeRefs(typeRefs: TypeRef[], forwardingRef?: TypeRef): TypeRef | undefined;
    finish(): TypeGraph;
    setLostTypeAttributes(): void;
    readonly lostTypeAttributes: boolean;
}
export interface UnionTypeProvider<TArrayData, TClassData, TMapData> {
    readonly arrayData: TArrayData;
    readonly mapData: TMapData;
    readonly classData: TClassData;
    enumCaseMap: {
        [name: string]: number;
    };
    enumCases: string[];
    getMemberKinds(): TypeAttributeMap<TypeKind>;
    readonly lostTypeAttributes: boolean;
}
export declare type TypeAttributeMap<T extends TypeKind> = OrderedMap<T, TypeAttributes>;
export declare class UnionAccumulator<TArray, TClass, TMap> implements UnionTypeProvider<TArray[], TClass[], TMap[]> {
    private readonly _conflateNumbers;
    private _nonStringTypeAttributes;
    private _stringTypeAttributes;
    readonly arrayData: TArray[];
    readonly mapData: TMap[];
    readonly classData: TClass[];
    private _lostTypeAttributes;
    enumCaseMap: {
        [name: string]: number;
    };
    enumCases: string[];
    constructor(_conflateNumbers: boolean);
    private have(kind);
    readonly haveString: boolean;
    addAny(attributes: TypeAttributes): void;
    addNull(attributes: TypeAttributes): void;
    addBool(attributes: TypeAttributes): void;
    addInteger(attributes: TypeAttributes): void;
    addDouble(attributes: TypeAttributes): void;
    addStringType(kind: PrimitiveStringTypeKind, attributes: TypeAttributes): void;
    addArray(t: TArray, attributes: TypeAttributes): void;
    addClass(t: TClass, attributes: TypeAttributes): void;
    addMap(t: TMap, attributes: TypeAttributes): void;
    addEnumCases(cases: OrderedMap<string, number>, attributes: TypeAttributes): void;
    addEnumCase(s: string, count: number, attributes: TypeAttributes): void;
    getMemberKinds(): TypeAttributeMap<TypeKind>;
    readonly lostTypeAttributes: boolean;
}
export declare class TypeRefUnionAccumulator extends UnionAccumulator<TypeRef, TypeRef, TypeRef> {
    private _typesAdded;
    private addType(t);
    private readonly numberOfAddedNonUnionTypes;
    addTypes(types: Set<Type>): TypeAttributes;
}
export declare abstract class UnionBuilder<TBuilder extends TypeBuilder, TArrayData, TClassData, TMapData> {
    protected readonly typeBuilder: TBuilder;
    constructor(typeBuilder: TBuilder);
    protected abstract makeEnum(cases: string[], counts: {
        [name: string]: number;
    }, typeAttributes: TypeAttributes, forwardingRef: TypeRef | undefined): TypeRef;
    protected abstract makeClass(classes: TClassData, maps: TMapData, typeAttributes: TypeAttributes, forwardingRef: TypeRef | undefined): TypeRef;
    protected abstract makeArray(arrays: TArrayData, typeAttributes: TypeAttributes, forwardingRef: TypeRef | undefined): TypeRef;
    private makeTypeOfKind(typeProvider, kind, typeAttributes, forwardingRef);
    buildUnion(typeProvider: UnionTypeProvider<TArrayData, TClassData, TMapData>, unique: boolean, typeAttributes: TypeAttributes, forwardingRef?: TypeRef): TypeRef;
}
